/*
 * This file is a part of PikoTools
 * and is distributed under the (new) BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2017-2021, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * Neither the name Tomasz Sowa nor the names of contributors to this
 *    project may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef headerfile_picotools_convert_strtoint
#define headerfile_picotools_convert_strtoint

#include <limits>
#include "text.h"
#include "misc.h"


namespace pt
{




template<typename CharType>
unsigned long long Toull(const CharType * str, int base = 10, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	unsigned long long res = 0;
	bool carry = false;
	int digit;

	SetOverflow(was_overflow, false);

	if( allow_skip_whitechars )
		str = skip_white(str);

	while( !carry && is_digit(*str, base, &digit) )
	{
	#ifdef __GNUC__
		carry = __builtin_mul_overflow(res, static_cast<unsigned long long>(base), &res);

		if( !carry )
		{
			carry = __builtin_add_overflow(res, static_cast<unsigned long long>(digit), &res);
		}
	#else
		// on other compilers than GCC or CLANG we do not test overflow at the moment
		res = res * static_cast<unsigned long long>(base) + static_cast<unsigned long long>(digit);
	#endif

		str += 1;
	}

	if( carry )
	{
		if( after_str )
		{
			while( is_digit(*str, base, &digit) )
			{
				str += 1;
			}
		}

		SetOverflow(was_overflow, true);
		res = 0;
	}

	if( after_str )
		*after_str = str;

	return res;
}


template<typename CharType>
long long Toll(const CharType * str, int base = 10, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	bool was_sign = false;
	bool was_overflow_u = false;

	SetOverflow(was_overflow, false);

	if( allow_skip_whitechars )
		str = skip_white(str);

	if( *str == '-' )
	{
		was_sign = true;
		str += 1;
	}

	unsigned long long uval = Toull(str, base, after_str, &was_overflow_u, false);
	unsigned long long sign_add = ( was_sign ) ? 1 : 0;

	if( was_overflow_u )
	{
		SetOverflow(was_overflow, true);
		return 0;
	}

	if( uval > static_cast<unsigned long long>(std::numeric_limits<long long>::max()) + sign_add )
	{
		SetOverflow(was_overflow, true);
		return 0;
	}

	if( was_sign )
	{
		return static_cast<long long>(0) - static_cast<long long>(uval);
	}

	return static_cast<long long>(uval);
}




template<typename CharType, typename IntegerType>
IntegerType ToUnsignedIntegerType(const CharType * str, int base = 10, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	bool was_overflow_ll = false;

	SetOverflow(was_overflow, false);
	unsigned long long val = Toull(str, base, after_str, &was_overflow_ll, allow_skip_whitechars);

	if( was_overflow_ll || val > static_cast<unsigned long long>(std::numeric_limits<IntegerType>::max()) )
	{
		SetOverflow(was_overflow, true);
		return 0;
	}

	return static_cast<IntegerType>(val);
}


template<class CharType>
unsigned long Toul(const CharType * str, int base = 10, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	return ToUnsignedIntegerType<CharType, unsigned long>(str, base, after_str, was_overflow, allow_skip_whitechars);
}

template<class CharType>
unsigned int Toui(const CharType * str, int base = 10, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	return ToUnsignedIntegerType<CharType, unsigned int>(str, base, after_str, was_overflow, allow_skip_whitechars);
}





template<typename CharType, typename IntegerType>
IntegerType ToIntegerType(const CharType * str, int base = 10, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	bool was_overflow_ll = false;

	SetOverflow(was_overflow, false);
	long long val = Toll(str, base, after_str, &was_overflow_ll, allow_skip_whitechars);

	if( was_overflow_ll ||
		val < static_cast<long long>(std::numeric_limits<IntegerType>::min()) ||
		val > static_cast<long long>(std::numeric_limits<IntegerType>::max()) )
	{
		SetOverflow(was_overflow, true);
		return 0;
	}

	return static_cast<IntegerType>(val);
}



template<class CharType>
long Tol(const CharType * str, int base = 10, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	return ToIntegerType<CharType, long>(str, base, after_str, was_overflow, allow_skip_whitechars);
}

template<class CharType>
int Toi(const CharType * str, int base = 10, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	return ToIntegerType<CharType, int>(str, base, after_str, was_overflow, allow_skip_whitechars);
}








/*
 *
 * the base will be automatically detected
 *
 * (first digit is [1-9]) - base 10
 * 0 - 8
 * # - 16
 * & - 2
 *
 */
template<typename CharType>
unsigned long long Toull_b(const CharType * str, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	if( allow_skip_whitechars )
		str = skip_white(str);

	int base = 10;

	if( *str == '0' )
	{
		base = 8;
		str += 1;
	}
	else
	if( *str == '#' )
	{
		base = 16;
		str += 1;
	}
	else
	if( *str == '&' )
	{
		base = 2;
		str += 1;
	}

	return Toull(str, base, after_str, was_overflow, false);
}



template<typename CharType>
long long Toll_b(const CharType * str, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	bool was_sign = false;
	bool was_overflow_u = false;

	SetOverflow(was_overflow, false);

	if( allow_skip_whitechars )
		str = skip_white(str);

	if( *str == '-' )
	{
		was_sign = true;
		str += 1;
	}

	unsigned long long uval = Toull_b(str, after_str, &was_overflow_u, false);
	unsigned long long sign_add = ( was_sign ) ? 1 : 0;

	if( was_overflow_u )
	{
		SetOverflow(was_overflow, true);
		return 0;
	}

	if( uval > static_cast<unsigned long long>(std::numeric_limits<long long>::max()) + sign_add )
	{
		SetOverflow(was_overflow, true);
		return 0;
	}

	if( was_sign )
	{
		return static_cast<long long>(0) - static_cast<long long>(uval);
	}

	return static_cast<long long>(uval);
}








template<typename CharType, typename IntegerType>
IntegerType ToUnsignedIntegerType_b(const CharType * str, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	bool was_overflow_ll = false;

	SetOverflow(was_overflow, false);
	unsigned long long val = Toull_b(str, after_str, &was_overflow_ll, allow_skip_whitechars);

	if( was_overflow_ll || val > static_cast<unsigned long long>(std::numeric_limits<IntegerType>::max()) )
	{
		SetOverflow(was_overflow, true);
		return 0;
	}

	return static_cast<IntegerType>(val);
}


template<class CharType>
unsigned long Toul_b(const CharType * str, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	return ToUnsignedIntegerType_b<CharType, unsigned long>(str, after_str, was_overflow, allow_skip_whitechars);
}

template<class CharType>
unsigned int Toui_b(const CharType * str, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	return ToUnsignedIntegerType_b<CharType, unsigned int>(str, after_str, was_overflow, allow_skip_whitechars);
}





template<typename CharType, typename IntegerType>
IntegerType ToIntegerType_b(const CharType * str, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	bool was_overflow_ll = false;

	SetOverflow(was_overflow, false);
	long long val = Toll_b(str, after_str, &was_overflow_ll, allow_skip_whitechars);

	if( was_overflow_ll ||
		val < static_cast<long long>(std::numeric_limits<IntegerType>::min()) ||
		val > static_cast<long long>(std::numeric_limits<IntegerType>::max()) )
	{
		SetOverflow(was_overflow, true);
		return 0;
	}

	return static_cast<IntegerType>(val);
}



template<class CharType>
long Tol_b(const CharType * str, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	return ToIntegerType_b<CharType, long>(str, after_str, was_overflow, allow_skip_whitechars);
}

template<class CharType>
int Toi_b(const CharType * str, const CharType ** after_str = 0, bool * was_overflow = 0, bool allow_skip_whitechars = true)
{
	return ToIntegerType_b<CharType, int>(str, after_str, was_overflow, allow_skip_whitechars);
}


}


#endif
