/*
 * This file is a part of PikoTools
 * and is distributed under the (new) BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2021, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * Neither the name Tomasz Sowa nor the names of contributors to this
 *    project may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef headerfile_picotools_convert_text_private
#define headerfile_picotools_convert_text_private

#include <string>
#include "text.h"


namespace pt
{

namespace pt_private
{

template<class CharType>
CharType to_lower_generic(CharType c)
{
	if( c >= 'A' && c <= 'Z' )
		return c - 'A' + 'a';

	return c;
}


template<class CharType>
CharType to_upper_generic(CharType c)
{
	if( c >= 'a' && c <= 'z' )
		return c - 'a' + 'A';

	return c;
}


template<class StringType>
void to_lower_str_generic(StringType & s)
{
	typename StringType::size_type i;

	for(i=0 ; i<s.size() ; ++i)
		s[i] = to_lower(s[i]);
}


template<class StringType>
void to_upper_str_generic(StringType & s)
{
	typename StringType::size_type i;

	for(i=0 ; i<s.size() ; ++i)
		s[i] = to_upper(s[i]);
}


template<class CharType>
CharType * skip_white_generic(CharType * str, bool check_additional_chars, bool treat_new_line_as_white)
{
	while( is_white(static_cast<wchar_t>(*str), check_additional_chars, treat_new_line_as_white) )
	{
		str += 1;
	}

	return str;
}


template<class CharType>
CharType * skip_white_from_back_generic(CharType * str_begin, CharType * str_end, bool check_additional_chars, bool treat_new_line_as_white)
{
	while( str_end > str_begin && is_white(static_cast<wchar_t>(*(str_end-1)), check_additional_chars, treat_new_line_as_white) )
	{
		str_end -= 1;
	}

	return str_end;
}


template<class CharType>
CharType * skip_white_from_back_generic(CharType * str, bool check_additional_chars, bool treat_new_line_as_white)
{
	CharType * str_begin = str;

	while( *str != 0 )
	{
		str += 1;
	}

	return skip_white_from_back_generic(str_begin, str, check_additional_chars, treat_new_line_as_white);
}




template<class StringType1, class StringType2>
int compare_generic(const StringType1 * str1, const StringType2 * str2)
{
	while( *str1 && *str2 && *str1 == *str2 )
	{
		++str1;
		++str2;
	}

	if( *str1 == 0 && *str2 == 0 )
		return 0;

	int c1;
	int c2;

	if constexpr (sizeof(StringType1) == 1 && sizeof(StringType2) == 1)
	{
		c1 = (wchar_t)(unsigned char)(*str1);
		c2 = (wchar_t)(unsigned char)(*str2);
	}
	else
	{
		c1 = *str1;
		c2 = *str2;
	}

	return c1 - c2;
}


template<class StringType1, class StringType2>
int compare_str_generic(const StringType1 & str1, const StringType2 & str2)
{
	return compare_generic(str1.c_str(), str2.c_str());
}


template<class StringType1, class StringType2>
int compare_generic(const StringType1 * str1_begin, const StringType1 * str1_end, const StringType2 * str2)
{
	while( str1_begin < str1_end && *str2 && *str1_begin == *str2 )
	{
		++str1_begin;
		++str2;
	}

	if( str1_begin == str1_end && *str2 == 0 )
		return 0;

	int c1;
	int c2;

	if constexpr (sizeof(StringType1) == 1 && sizeof(StringType2) == 1)
	{
		c1 = str1_begin < str1_end ? (wchar_t)(unsigned char)(*str1_begin) : 0;
		c2 = (wchar_t)(unsigned char)(*str2);
	}
	else
	{
		c1 = str1_begin < str1_end ? *str1_begin : 0;
		c2 = *str2;
	}

	return c1 - c2;
}



template<class StringType1, class StringType2>
int compare_nc_generic(const StringType1 * str1, const StringType2 * str2)
{
	while( *str1 && *str2 && to_lower(*str1) == to_lower(*str2) )
	{
		++str1;
		++str2;
	}

	if( *str1 == 0 && *str2 == 0 )
		return 0;

	int c1;
	int c2;

	if constexpr (sizeof(StringType1) == 1 && sizeof(StringType2) == 1)
	{
		c1 = to_lower((wchar_t)(unsigned char)(*str1));
		c2 = to_lower((wchar_t)(unsigned char)(*str2));
	}
	else
	{
		c1 = to_lower(*str1);
		c2 = to_lower(*str2);
	}

	return c1 - c2;
}


template<class StringType1, class StringType2>
int compare_nc_str_generic(const StringType1 & str1, const StringType2 & str2)
{
	return compare_nc(str1.c_str(), str2.c_str());
}


template<class StringType1, class StringType2>
int compare_nc_generic(const StringType1 * str1_begin, const StringType1 * str1_end, const StringType2 * str2)
{
	while( str1_begin < str1_end && *str2 && to_lower(*str1_begin) == to_lower(*str2) )
	{
		++str1_begin;
		++str2;
	}

	if( str1_begin == str1_end && *str2 == 0 )
		return 0;

	int c1;
	int c2;

	if constexpr (sizeof(StringType1) == 1 && sizeof(StringType2) == 1)
	{
		c1 = str1_begin < str1_end ? to_lower((wchar_t)(unsigned char)(*str1_begin)) : 0;
		c2 = to_lower((wchar_t)(unsigned char)(*str2));
	}
	else
	{
		c1 = str1_begin < str1_end ? to_lower(*str1_begin) : 0;
		c2 = to_lower(*str2);
	}

	return c1 - c2;
}




template<class StringType1, class StringType2>
bool is_substr_generic(const StringType1 * short_str, const StringType2 * long_str)
{
	while( *short_str && *long_str && *short_str == *long_str )
	{
		++short_str;
		++long_str;
	}

	if( *short_str == 0 )
		return true;

	return false;
}



template<class StringType1, class StringType2>
bool is_substr_nc_generic(const StringType1 * short_str, const StringType2 * long_str)
{
	while( *short_str && *long_str && to_lower(*short_str) == to_lower(*long_str) )
	{
		++short_str;
		++long_str;
	}

	if( *short_str == 0 )
		return true;

	return false;
}




} // namespace pt_private

} // namespace pt


#endif

