/*
 * This file is a part of PikoTools
 * and is distributed under the (new) BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2021, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *
 *  * Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 *  * Neither the name Tomasz Sowa nor the names of contributors to this
 *    project may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef headerfile_picotools_csv_csvparser
#define headerfile_picotools_csv_csvparser

#include "space/space.h"
#include <string>
#include <fstream>


namespace pt
{

/*
 *
 * https://datatracker.ietf.org/doc/html/rfc4180
 *
 */
class CSVParser
{
public:

	enum Status
	{
		ok,
		cant_open_file,
	};


	Status parse_file(const char * file_name,         Space & out_space);
	Status parse_file(const std::string & file_name,  Space & out_space);
	Status parse_file(const wchar_t * file_name,      Space & out_space);
	Status parse_file(const std::wstring & file_name, Space & out_space);

	Status parse(const char * str,                    Space & out_space);
	Status parse(const std::string & str,             Space & out_space);
	Status parse(const wchar_t * str,                 Space & out_space);
	Status parse(const std::wstring & str,            Space & out_space);



protected:


	/*
		the last status of parsing, set by Parse() methods
	*/
	Status status;


	Space * space;

	/*
		true if parse_file() method was called
		false if parse() was called
	*/
	bool reading_from_file;

	/*
		true if parse(wchar_t *) or parse(std::wstring&) was called
	*/
	bool reading_from_wchar_string;

	/*
		pointers to the current character
		if parse() is being used
	*/
	const char    * pchar_ascii;
	const wchar_t * pchar_unicode;


	/*
		last read char
		or -1 if the end
	*/
	int lastc;



	/*
		a number of a line in which there is a syntax_error
	*/
	int line;

	/*
		current file

		may it would be better to make a pointer?
		if we parse only a string then there is no sense to have such an object
	*/
	std::ifstream file;

	/*
		input file is in UTF-8
		default: true
	*/
	bool input_as_utf8;




	void parse();
	void parse_row(Space & row_space);

	bool read_value_to(Space & row_space);
	bool read_quoted_value_to(std::wstring & value);
	bool read_non_quoted_value_to(std::wstring & value);



	/*
	 * copied from SpaceParser
	 * may it would be better to have a class with those methods and inherit from it?
	 */
	int read_utf8_char();
	int read_ascii_char();
	int read_char_from_wchar_string();
	int read_char_from_utf8_string();
	int read_char_from_ascii_string();
	int read_char_no_escape();

	int read_char();
};

}

#endif
